package org.processmining.ptconversions.epc;

import java.util.HashMap;
import java.util.Map;

import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.models.graphbased.directed.epc.ConfigurableEPC;
import org.processmining.models.graphbased.directed.epc.EPCFactory;
import org.processmining.models.graphbased.directed.epc.EPCNode;
import org.processmining.models.graphbased.directed.epc.elements.Connector;
import org.processmining.models.graphbased.directed.epc.elements.Connector.ConnectorType;
import org.processmining.models.graphbased.directed.epc.elements.Event;
import org.processmining.models.graphbased.directed.epc.elements.Function;
import org.processmining.processtree.Block;
import org.processmining.processtree.Block.Seq;
import org.processmining.processtree.Event.Message;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Task.Manual;
import org.processmining.processtree.impl.AbstractTask.Automatic;
import org.processmining.processtree.impl.ProcessTreeImpl;

/**
 * 
 * @author DSchunse
 * TODO: configurability
 */
public class ProcessTree2EPC {
	Map<Node, Pair<EPCNode, EPCNode>> transformation;
	ProcessTree tree;
	ConfigurableEPC cepc;
	
	public ConfigurableEPC convert(PluginContext context, ProcessTree tree, boolean unfoldTree){
		this.transformation = new HashMap<Node, Pair<EPCNode,EPCNode>>();
		this.tree = new ProcessTreeImpl(tree);
		if(unfoldTree){
			this.tree = this.tree.toTree();
		}
		cepc = EPCFactory.newConfigurableEPC(this.tree.getName());
		Pair<EPCNode, EPCNode> fragment = convertNode(this.tree.getRoot());
		if(fragment.equals(new Pair<EPCNode, EPCNode>(null, null))){
			Event startEvent = cepc.addEvent("Start event");
			Function startFunction = cepc.addFunction("Start function");
			Event endEvent = cepc.addEvent("End event");
			cepc.addArc(startEvent, startFunction);
			cepc.addArc(startFunction, endEvent);
		}
		else{
			// if the root is not an seq of a task, then we have to introduce some extra stuff
			if(this.tree.getRoot() instanceof Manual){
				// don't bother
			}
			else if(this.tree.getRoot() instanceof Seq){
				// check if the left most child is an event
				Node leftMostChild = ((Seq)this.tree.getRoot()).getChildren().get(0);
				while(leftMostChild instanceof Seq){
					leftMostChild = ((Seq)leftMostChild).getChildren().get(0);
				}
				// we now have the left most child
				if(leftMostChild instanceof Manual){
					// we are done
				}
				else{
					Event startEvent = cepc.addEvent("Start event");
					Function startFunction = cepc.addFunction("Start function");
					cepc.addArc(startEvent, startFunction);
					cepc.addArc(startFunction, fragment.getFirst());
				}
			}
			else{
				Event startEvent = cepc.addEvent("Start event");
				Function startFunction = cepc.addFunction("Start function");
				cepc.addArc(startEvent, startFunction);
				cepc.addArc(startFunction, fragment.getFirst());
			}
			Event endEvent = cepc.addEvent("End event");
			cepc.addArc(fragment.getSecond(), endEvent);
		}
		return cepc;
		
	}
	
	private Pair<EPCNode, EPCNode> convertNode(Node n){
		if(transformation.containsKey(n)){
			return transformation.get(n);
		}
		Pair<EPCNode, EPCNode> fragment = null;
		switch(tree.getType(n)){
			case AND:
				fragment = convertANDXORORDEFPLACEHOLDER((Block)n);
				break;
			case AUTOTASK:
				fragment = convertAutomatic((Automatic)n);
				break;
			case DEF:
				fragment = convertANDXORORDEFPLACEHOLDER((Block)n);
				break;
			case LOOPDEF:
				fragment = convertLoop((Block)n);
				break;
			case LOOPXOR:
				fragment = convertLoop((Block)n);
				break;
			case MANTASK:
				fragment = convertManual((Manual) n);
				break;
			case MESSAGE:
				fragment = convertEvent((org.processmining.processtree.Event)n);
				break;
			case OR:
				fragment = convertANDXORORDEFPLACEHOLDER((Block)n);
				break;
			case PLACEHOLDER:
				fragment = convertANDXORORDEFPLACEHOLDER((Block)n);
				break;
			case SEQ:
				fragment = convertSEQ((Seq) n);
				break;
			case TIMEOUT:
				fragment = convertEvent((org.processmining.processtree.Event)n);
				break;
			case XOR:
				fragment = convertANDXORORDEFPLACEHOLDER((Block)n);
				break;
			default:
				return null;
		}
		if(fragment.equals(new Pair<EPCNode, EPCNode>(null, null))){
			transformation.put(n, fragment);
			return fragment;
		}
		if(n.getIncomingEdges().size() > 1){
			// if we have incoming edges, we need to put some xors before and after
			Connector XORJoin = cepc.addConnector("", ConnectorType.XOR);
			Connector XORSplit = cepc.addConnector("", ConnectorType.XOR);
			cepc.addArc(XORJoin, fragment.getFirst());
			cepc.addArc(fragment.getSecond(), XORSplit);
			Pair<EPCNode, EPCNode> newFragment = new Pair<EPCNode, EPCNode>(XORJoin, XORSplit);
			transformation.put(n, newFragment);
			return newFragment;
		}
		else{
			transformation.put(n, fragment);
			return fragment;
		}
	}
	
	private Pair<EPCNode, EPCNode> convertAutomatic(Automatic n){
		Pair<EPCNode, EPCNode> fragment = new Pair<EPCNode, EPCNode>(null, null);
		return fragment;
	}
	
	private Pair<EPCNode, EPCNode> convertManual(Manual n){
		Event event = cepc.addEvent("Event" + n.getName());
		Function function = cepc.addFunction(n.getName());
		cepc.addArc(event, function);
		Pair<EPCNode, EPCNode> fragment = new Pair<EPCNode, EPCNode>(event, function);
		return fragment;
	}
	
	private Pair<EPCNode, EPCNode> convertANDXORORDEFPLACEHOLDER(Block n){
		Connector startBlock = cepc.addConnector("Split" + n.getName(), convertType(n));
		Connector endBlock = cepc.addConnector("Join", convertType(n));
		boolean startToEnd = false;
		for(Node child: n.getChildren()){
			Pair<EPCNode, EPCNode> childFragment = convertNode(child);
			if(childFragment.equals(new Pair<EPCNode, EPCNode>(null, null))){
				if(!startToEnd){
					cepc.addArc(startBlock, endBlock);
					startToEnd = true;
				}
			}
			else{
				Pair<EPCNode, EPCNode> startEnd = convertNode(child);
				cepc.addArc(startBlock, startEnd.getFirst());
				cepc.addArc(startEnd.getSecond(), endBlock);
			}
		}
		Pair<EPCNode, EPCNode> fragment = new Pair<EPCNode, EPCNode>(startBlock, endBlock);
		return fragment;
	}
	
	private Pair<EPCNode, EPCNode> convertSEQ(Seq n){
		EPCNode startNode = null;
		EPCNode endNode = null;
		Pair<EPCNode, EPCNode> previousFragment = new Pair<EPCNode, EPCNode>(null, null);
		for(int i = 0; i < n.getChildren().size(); i++){
			Pair<EPCNode, EPCNode> fragment = convertNode(n.getChildren().get(i));
			if(!fragment.equals(new Pair<EPCNode, EPCNode>(null, null))){
				if(startNode == null){
					startNode = fragment.getFirst();
				}
				else{
					cepc.addArc(previousFragment.getSecond(), fragment.getFirst());
				}
				previousFragment = fragment;
			}
		}
		endNode = previousFragment.getSecond();
		Pair<EPCNode, EPCNode> fragment = new Pair<EPCNode, EPCNode>(startNode, endNode);
		return fragment;
	}
	
	private Pair<EPCNode, EPCNode> convertEvent(org.processmining.processtree.Event n){
		Event event;
		Function function;
		if(n instanceof Message){
			event = cepc.addEvent("Event Message " + n.getName());
			function = cepc.addFunction("Message " + n.getName());
		}
		else{
			event = cepc.addEvent("Event Time-Out " + n.getName());
			function = cepc.addFunction("Time-Out " + n.getName());
		}
		cepc.addArc(event, function);
		Pair<EPCNode, EPCNode> childFragment = convertNode(n.getChildren().get(0));
		Pair<EPCNode, EPCNode> fragment = null;
		if(childFragment.equals(new Pair<EPCNode, EPCNode>(null, null))){
			fragment = new Pair<EPCNode, EPCNode>(event, function);
		}
		else{
			cepc.addArc(function, childFragment.getFirst());
			fragment = new Pair<EPCNode, EPCNode>(event, childFragment.getSecond());
		}
		return fragment;
	}
	
	private Pair<EPCNode, EPCNode> convertLoop(Block n){
		Connector XORJoin = cepc.addConnector("Join" + n.getName(), ConnectorType.XOR);
		Connector XORSplit = cepc.addConnector("Split" + n.getName(), ConnectorType.XOR);
		Pair<EPCNode, EPCNode> fragmentDo = convertNode(n.getChildren().get(0));
		Pair<EPCNode, EPCNode> fragmentRedo = convertNode(n.getChildren().get(1));
		Pair<EPCNode, EPCNode> fragmentExit = convertNode(n.getChildren().get(2));
		if(fragmentDo.equals(new Pair<EPCNode, EPCNode>(null, null))){
			cepc.addArc(XORJoin, XORSplit);
		}
		else{
			cepc.addArc(XORJoin, fragmentDo.getFirst());
			cepc.addArc(fragmentDo.getSecond(), XORSplit);
		}
		if(fragmentRedo.equals(new Pair<EPCNode, EPCNode>(null, null))){
			cepc.addArc(XORSplit, XORJoin);
		}
		else{
			cepc.addArc(XORSplit, fragmentRedo.getFirst());
			cepc.addArc(fragmentRedo.getSecond(), XORJoin);
		}
		Pair<EPCNode, EPCNode> fragment = null;
		if(fragmentExit.equals(new Pair<EPCNode, EPCNode>(null, null))){
			fragment = new Pair<EPCNode, EPCNode>(XORJoin, XORSplit);
		}
		else{
			cepc.addArc(XORSplit, fragmentExit.getFirst());
			fragment = new Pair<EPCNode, EPCNode>(XORJoin, fragmentExit.getSecond());
		}
		return fragment;
	}
	
	private ConnectorType convertType(Block n){
		switch(tree.getType(n)){
			case AND:
				return ConnectorType.AND;
			case XOR:
				return ConnectorType.XOR;
			case OR:
				return ConnectorType.OR;
			case DEF:
				return ConnectorType.XOR;
			case PLACEHOLDER:
				return ConnectorType.XOR;
			default:
				throw new NullPointerException();
		}
	}
}
